package App::GraphQL;

use strict;
use warnings;
use utf8;
use open qw/:utf8/;

use App::GraphQL::Cache;
use App::GraphQL::Context;
use App::GraphQL::ErrorHandler;
use App::GraphQL::Repository::Query;
use App::GraphQL::Repository::Request;
use Class::Load qw(try_load_class);
use File::Slurp;
use GraphQL::Schema;
use GraphQL::Language::Parser;
use GraphQL::Execution;
use Promise::XS;
use Scalar::Util qw(blessed);

sub execute {
    # $query にはリクエストされたクエリが渡されます
    # $variable…
    # $operation_name…
    my ($class, $query, $variables, $operation_name) = @_;

    # ここでスキーマを読み込みパースします
    my $schema_file = read_file('schema.graphql', binmode => ':utf8');
    my $schema = GraphQL::Schema->from_doc(
        $schema_file,
        +{
            %GraphQL::Schema::KIND2CLASS,
            # interface で型解決を個別に移譲できるように拡張する
            interface => 'App::GraphQL::Type::Interface',
        },
    );

    # ここでクエリをパースします
    my $parsed_query = GraphQL::Language::Parser::parse($query);

    # Mutation ではキャッシュしない
    my $cachable = !!0;
    for my $query ($parsed_query->@*) {
        if (($_->{operationType} // '') eq 'mutation' && ($_->{name} // '') eq $operation_name) {
            $cachable = !!1;
        }
    }

    # Context を作成する．これは各 Resolver で参照できるため，共通のストアを持つこともできる
    my $context = App::GraphQL::Context->new();

    # スキーマとクエリを基に
    my $result = GraphQL::Execution::execute(
        $schema,
        $parsed_query,
        {},
        $context,
        $variables,
        $operation_name,
        \&_resolver_with_error_handler,
        +{
            resolve => \&Promise::XS::resolved,
            reject  => \&Promise::XS::rejected,
            all     => \&_promise_all,
        },
    );

    # リクエストが完了したタイミングで DataLoader の遅延評価を行う
    Promise::XS::resolved()->then(sub { $result })->then(sub { $result = shift });
    $context->dispatch_data_loaders();

    return ($result, $cachable);
}

sub _promise_all {
    return Promise::XS::all(map {
        (blessed($_) && $_->can('then'))
            ? $_ : Promise::XS::resolved($_)
    } @_);
}

sub execute_with_cache {
    my ($class, $payload) = @_;
    my $query = $payload->{query} // '';
    my $variables = $payload->{varaibles} // +{};
    my $operation_name = $payload->{operationName} // '';
    my $cached_response = App::GraphQL::Cache->fetch_cache($payload);
    if (defined $cached_response) {
        return $cached_response;
    }
    my ($response, $cachable) = $class->execute($query, $variables, $operation_name);
    if ($cachable) {
        App::GraphQL::Cache->set_cache($payload, $response);
        App::GraphQL::Repository::Request->save($query, $variables, $operation_name);
    }
    $class->_persist($payload);

    return $response;
}

sub execute_persisted_query {
    my ($class, $operation_name, $variables, $extensions) = @_;
    my $hash = $extensions->{persistedQuery}->{sha256Hash} // '';
    if (my $persisted_query = App::GraphQL::Repository::Query->restore($hash)) {
        return $class->execute_with_cache(+{
            operationName => $operation_name,
            query         => $persisted_query,
            variables     => $variables,
        });
    }
    return +{errors => [+{
        message    => 'No Persisted Query Available',
        extensions => +{ code => 'PERSISTED_QUERY_NOT_FOUND' },
    }]};
}

sub _resolver_with_error_handler {
    my ($root_value, $args, $context, $info) = @_;
    my $result = eval { _resolver(@_) };
    my $error = $@;

    if ($error) {
        return App::GraphQL::ErrorHandler->handle($error);
    } else {
        return $result;
    }
}

# 型の Resolver
sub _resolver {
    my ($root_value, $args, $context, $info) = @_;
    # $root_value: 対象のスコープの型
    # $args:       フィールドに渡されている引数のHashRef
    # $context:    GraphQL::Execution::execute の第4引数に渡されたオブジェクト
    # $info:       型の情報やスキーマのASTなどの情報

    # 対象のフィールドの名前が得られる
    my $field_name = $info->{field_name};

    # 単純に HashRef のフィールドであれば，その値をそのまま返す
    my $is_hashref = (!blessed($root_value) && ref $root_value eq 'HASH');
    if ($is_hashref && exists $root_value->{$field_name}) {
        return $root_value->{$field_name};
    }

    # 個別のResolverモジュールに移譲する
    my $parent_name = $info->{parent_type}->name;
    my $resolver_impl = join '::', (__PACKAGE__, 'Resolver', $parent_name);
    my ($resolver_loaded) = try_load_class($resolver_impl);
    if ($resolver_loaded && $resolver_impl->can($field_name)) {
        return $resolver_impl->$field_name($root_value, $args, $context, $info);
    }

    # 解決できずなにも得られなかった！
    return undef;
}

sub _persist {
    my ($class, $payload) = @_;
    my $query = $payload->{query};
    my $extensions_decoded = $payload->{extensions} // +{};
    my $persisted_query_info = $extensions_decoded->{persistedQuery} // +{};
    my $hash = $persisted_query_info->{sha256Hash};
    if ($hash && $query) {
        App::GraphQL::Repository::Query->save($hash, $query);
    }
}

1;
